"""Preconfigured converters for msgpack."""
from cattrs.preconf.msgpack import configure_converter, make_converter, MsgpackConverter

__all__ = ["configure_converter", "make_converter", "MsgpackConverter"]
