"""Preconfigured converters for orjson."""
from cattrs.preconf.orjson import configure_converter, make_converter, OrjsonConverter

__all__ = ["configure_converter", "make_converter", "OrjsonConverter"]
